/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dpop;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author Andrea
 */
public class Handler<K, D> {
    private TaggedDictionary<IntegerValue, GraphNode> _nodes;
    private ArrayList<PseudoTreeNode> _isolated_nodes;
    private TaggedDictionary<PseudoTreeNode, Value> _solution;
    private Value _solution_weight;
    private TaggedDictionary<PseudoTreeNode, ArrayList<PseudoTreeNode>> _broken_constraints;

    private Interface _view;
    private boolean _processing;
    private boolean _use_precise_timing;
    private String _file_read;

    //the most connected node of the graph read, from which to start the dfs with the 
    //heuristic MCN
    private GraphNode _most_connected_node;
    //the first node of the graph read, to use without heuristic
    private GraphNode _first_node;
    private PseudoTreeNode _root;
    
    Handler(Interface view) {
        _view = view;
        
        _initialize_solution();
        _processing = false;

        _nodes = null;
        _first_node = null;
        _most_connected_node = null;
    }

    // --------------------- USER INTERFACE HANDLING PART ---------------------

    public void read_input_file(String file, boolean heuristic) {
        //handling UI
        _view.clear();
        _view.setSourceGraphEnabled(true);
        _view.setSourcePseudoTreeEnabled(true);
        _view.setStartEnabled(true);

        //parsing file
        boolean res = _parse_graph_coloring(file);

        if (res) {
            _file_read = file;

            _view.addLogLine("Parsed file " + new File(_file_read).getName() + ".");

            //now starts the conversion of the graph to the pseudo tree
            if (heuristic) {
                _root = _most_connected_node.dfs_visit_mcn();
            }
            else {
                _root = _first_node.dfs_visit();
            }

            _view.addLogLine("PseudoTree created.");

            //finally, inspect if there are isolated nodes
            _inspect_isolated_nodes();
        }
    }

    private void _inspect_isolated_nodes() {
        _isolated_nodes = new ArrayList<PseudoTreeNode>();

        Set<Map.Entry<IntegerValue, GraphNode>> s = _nodes.entrySet();
        Iterator<Map.Entry<IntegerValue, GraphNode>> it = s.iterator();
        while (it.hasNext()) {
            Map.Entry<IntegerValue, GraphNode> e = it.next();
            GraphNode current = e.getValue();
            if (current.isIsolated()) {
                //creates the pseudotreenode associated
                PseudoTreeNode n = new PseudoTreeNode(current, this);

                _isolated_nodes.add(n);
            }
        }
    }

    private void _estimate_calculation() {
        _view.setStateLabel("Calculating..");
        _use_precise_timing = _view.isPreciseTiming();

        int uw = _root.evaluate_util_weight();
        int total = uw + (_nodes.size() - 1);

        _view.setProgress((int)total);
    }

    public void start_stop() {
        if (!_processing) {
            //start
            _processing = true;

            _view.start();

            _view.addLogLine("Begin calculation on file " + new File(_file_read).getName() + "\n");

            _estimate_calculation();
            _initialize_solution();

            _start_computation();
        }
        else {
            //stop
            _processing = false;
            _view.stop();
        }
    }

    public void end() {
        _processing = false;
        _view.halted();
    }

    public boolean isProcessing() {
        return _processing;
    }

    public boolean usePreciseTiming() {
        return _use_precise_timing;
    }

    // ---------------------- PROCESS HANDLING PART -----------------------

    private void _initialize_solution() {
        _solution = new TaggedDictionary<PseudoTreeNode, Value>(this);
        _solution_weight = null;
        _broken_constraints = new TaggedDictionary<PseudoTreeNode, ArrayList<PseudoTreeNode>>(this);
    }

    private boolean _parse_graph_coloring(String file) {
        try {

            BufferedReader instream = null;
            String line = null;
            StringTokenizer t;
            BinaryConstraint current_bc = null;
            GraphNode current_n1 = null;
            GraphNode current_n2 = null;
            _most_connected_node = null;
            _first_node = null;
            _nodes = new TaggedDictionary<IntegerValue, GraphNode>();

            boolean end = false;

            instream = new BufferedReader(new FileReader(file));

            while (!end) {
                line = instream.readLine();

                if (line == null) {
                    //end of file
                    break;
                }

                t = new StringTokenizer(line);
                String command = t.nextToken();

                if (command.compareToIgnoreCase("variable") == 0) {
                    IntegerValue id = new IntegerValue(new Integer(t.nextToken()).intValue());
                    t.nextToken();  //skip the agent
                    int domain_size = new Integer(t.nextToken()).intValue();

                    //creates the node from this variable line
                    GraphNode n = new GraphNode(id, _create_domain_from_size(domain_size), this);
                    _nodes.put(id, n);
                }
                else if (command.compareToIgnoreCase("constraint") == 0) {
                    IntegerValue v1 = new IntegerValue(new Integer(t.nextToken()).intValue());
                    IntegerValue v2 = new IntegerValue(new Integer(t.nextToken()).intValue());

                    //searches for the 2 variables v1 and v2 in the nodes list
                    current_n1 = _nodes.get(v1);
                    current_n2 = _nodes.get(v2);
                    current_bc = new BinaryConstraint(current_n1, current_n2);

                    current_n1.add_neighbor(current_n2, current_bc);

                    //keep the most connected node reference updated
                    GraphNode selected;
                    if (current_n1.getNeighborsCount() > current_n2.getNeighborsCount()) {
                        selected = current_n1;
                    }
                    else {
                        selected = current_n2;
                    }
                    if (_most_connected_node == null || selected.getNeighborsCount() > _most_connected_node.getNeighborsCount()) {
                        _most_connected_node = selected;
                    }
                }
                else if (command.compareToIgnoreCase("nogood") == 0) {
                    IntegerValue v1 = new IntegerValue(new Integer(t.nextToken()).intValue());
                    IntegerValue v2 = new IntegerValue(new Integer(t.nextToken()).intValue());
                    //IntegerValue cost = new IntegerValue(new Integer(t.nextToken()));
                    IntegerValue cost = new IntegerValue(-1);

                    Couple c1 = new Couple(current_n1, v1);
                    Couple c2 = new Couple(current_n2, v2);

                    current_bc.addConstraint(c1, c2, cost);
                }
            }

            if (_most_connected_node == null) {
                _most_connected_node = (_nodes.size() > 0) ? (GraphNode)_nodes.values().iterator().next() : null;
            }
            _first_node = (_nodes.size() > 0) ? (GraphNode)_nodes.values().iterator().next() : null;

            return true;
        }
        catch (Exception e) {
            javax.swing.JOptionPane.showMessageDialog(null, "Error reading input file", "Error", javax.swing.JOptionPane.ERROR_MESSAGE);
            return false;
        }
    }

    public GraphNode getGraphRoot() {
        return _most_connected_node;
    }

    public PseudoTreeNode getPseudoTreeRoot() {
        return _root;
    }

    private ArrayList<Value> _create_domain_from_size(int size) {
        int i = 0;
        ArrayList<Value> res = new ArrayList<Value>();
        while (i < size) {
            Value current = new IntegerValue(i);
            res.add(current);

            i++;
        }

        return res;
    }

    private void _start_computation() {
        _root.start();

        //do the start also to eventual isolated nodes
        Iterator<PseudoTreeNode> it = _isolated_nodes.iterator();
        while (it.hasNext()) {
            it.next().start();
        }
    }

    public void evaluating_util(PseudoTreeNode source, ArrayList<PseudoTreeNode> variables) {
        String variables_list = "";
        Iterator<PseudoTreeNode> it = variables.iterator();
        boolean first = true;
        while (it.hasNext()) {
            if (!first) {
                variables_list = variables_list + ", ";
            }
            variables_list = variables_list + it.next().shortToString();
            first = false;
        }
        _view.addLogLine(source.shortToString() + " is evaluating util between variables " + variables_list);
    }

    public void incrementProgress(int v) {
        _view.incrementProgress(v);
    }

    public void util_sent(PseudoTreeNode source, PseudoTreeNode target) {
        _view.addLogLine(source.shortToString() + " has sent util to " + target.shortToString());
    }

    public void value_sent(PseudoTreeNode source, PseudoTreeNode target) {
        _view.incrementProgress(1);

        if (target == null) {
            _view.addLogLine(source.shortToString() + " (isolated node) has comunicated its value");
        }
        else {
            _view.addLogLine(source.shortToString() + " has sent value to " + target.shortToString());
        }
    }

    public void broken_constraint(PseudoTreeNode n1, PseudoTreeNode n2) {
        if (_broken_constraints.containsKey(n1)) {
            ArrayList<PseudoTreeNode> arr = _broken_constraints.get(n1);
            arr.add(n2);
        }
        else {
            ArrayList<PseudoTreeNode> arr = new ArrayList<PseudoTreeNode>();
            arr.add(n2);
             _broken_constraints.put(n1, arr);
        }
    }

    public void receive_solution(PseudoTreeNode n, Value v) {
        _solution.put(n, v);

        if (_solution.size() == _nodes.size()) {
            _view.show_solution(_solution, _solution_weight, _broken_constraints);
        }
    }

    public void receive_solution_weight(Value w) {
        _solution_weight = w;
    }
}
